/*
 *  hnv_closure.c
 *  HeteroNumberViewer
 *
 *  Created by Naoto Morikawa on 9/17/07.
 *  Copyright (C) 2007 Naoto Morikawa. All rights reserved.
 *
 *  2007-10-01 version 0.1 release
 */

#include <stdio.h>
#include <stdlib.h>
#include "hnv_main.h"
#include "hnv_computation_main.h"
#include "hnv_cone.h"
#include "hnv_tetrahedron.h"
#include "hnv_trajectory.h"
#include "hnv_set_up_tetra_flow.h"

struct peak_node_chunk {
	int array_size;
	struct peak_node * node_array;
	int chunk_ct;
	struct peak_node_chunk * next;
};


struct peak_node {
	struct hnv_coor_4d coor;
	
	struct peak_node * next;
	struct peak_node * left;
	struct peak_node * right;
};


//===============================================
// Private data
//===============================================

struct peak_node_chunk peak_node_pool_head;
struct peak_node unused_peak_node_head;
struct peak_node bd_peak_list_head;
struct peak_node * peak_tree_root = NULL;

// Neighbours around peak p={ 0, 0, 0 }
float neighbours[][3] = {
	{  2.0,  0.0,  2.0 },		//     x/y
	{  2.0,  2.0,  0.0 },		//     x/z
	{  0.0,  2.0,  2.0 },		//     x/w
	{ -2.0,  0.0, -2.0 },		//     y/x
	{  0.0,  2.0, -2.0 },		//     y/z
	{ -2.0,  2.0,  0.0 },		//     y/w
	{ -2.0, -2.0,  0.0 },		//     z/x
	{  0.0, -2.0,  2.0 },		//     z/y
	{ -2.0,  0.0,  2.0 },		//     z/w
	{  0.0, -2.0, -2.0 },		//     w/x
	{  2.0, -2.0,  0.0 },		//     w/y
	{  2.0,  0.0, -2.0 }		//     w/z
};

// bottom points of tetrahedrons around peak p={ 0, 0, 0 }
float bottoms[][3] = {
	{ -1.0,  1.0,  1.0 },		//     xyz 
	{  1.0,  1.0, -1.0 },		//     xyw
	{  1.0, -1.0,  1.0 },		//     xzw 
	{ -1.0, -1.0, -1.0 }		//     yzw 
};


//===============================================
// Public data
//===============================================

//===============================================
// Private methods
//===============================================
//
// void initialize_pnode_chunk(struct pnode_chunk * new_chunk);
// void put_pnode_chunk(struct peak_node_chunk * list_head, 
//			struct peak_node_chunk * new_chunk);
// void put_peak_node(struct peak_node * list_head,
//		      struct peak_node * pnode);
// struct peak_node * take_peak_node(struct peak_node * list_head);
// int allocate_peak_nodes(void);
// void initialize_peak_node(struct peak_node * pnode);
// void set_up_empty_pnode_list(struct peak_node * list_head);
// void fill_peak_node_list(struct peak_node * list_head);
// struct peak_node * assign_peak_node(void);
// void recall_peak_node(struct peak_node * pnode);
// HNV_COMPARE_RESULT which_peak_is_greater(struct peak_node * current_node,
//					      struct peak_node * new_node);
// void print_peak_node_chunk(struct peak_node_chunk * p_chunk);
// void print_pnode_chunk_list(struct peak_node_chunk * chunk);
// void print_peak_node(struct peak_node * pnode);
// void walk_peak_node_tree(struct peak_node * pnode);
// void print_peak_node_tree(void);
// void print_out_new_peaks(FILE * stream, struct peak_node * peak);
// void print_pnode_list(struct peak_node * peak);
// void initialize_peak_node_tree(void);
// void finalize_peak_node_tree(void);
// void register_in_peak_tree(struct peak_node * new_pnode);
// struct peak_node * check_if_peak_is_processed(struct peak_node * new_pnode);
// void set_new_peak(struct peak_node * new_peak, struct hnv_coor_3d * pt_bottom,
//		     float height);
// void compute_top_vertex(struct hnv_coor_3d * vertex, struct peak_node * peak,
// 			   int top_ind);
// void compute_bottom_vertex(struct hnv_coor_3d * bottom, struct hnv_coor_3d * top,
//			      int bottom_ind);
// int check_if_on_boundary(struct hnv_coor_3d * top, struct hnv_coor_3d * bottom, 
//			    struct hnv_cone * std_cone, struct hnv_cone * conj_cone);
// int compute_bd_peak_list(struct hnv_cone * std_cone, struct hnv_cone * conj_cone);
// struct hnv_cone * closure_of_std_cone(struct hnv_cone * std_cone, struct hnv_cone * conj_cone);
//-----------------------------------------------

void initialize_pnode_chunk(struct peak_node_chunk * new_chunk)
{
	new_chunk->array_size = 0;
	new_chunk->node_array = NULL;
	new_chunk->chunk_ct = 0;
	new_chunk->next = NULL;
	return;
}


void put_pnode_chunk(struct peak_node_chunk * list_head, 
		     struct peak_node_chunk * new_chunk)
{
	new_chunk->next = list_head->next;
	list_head->next = new_chunk;
	list_head->chunk_ct ++;
	return;
}


void put_peak_node(struct peak_node * list_head,
		     struct peak_node * pnode)
{
	pnode->next = list_head->next;
	list_head->next = pnode;
	return;
}


struct peak_node * take_peak_node(struct peak_node * list_head)
{
	struct peak_node * pnode;
	
	pnode = list_head->next;
	if (pnode == NULL)
		return NULL;
	list_head->next = pnode->next;
	pnode->next = NULL;
	return pnode;
}


void initialize_peak_node(struct peak_node * pnode)
{
	pnode->coor.k = 0.0;
	pnode->coor.l = 0.0;
	pnode->coor.m = 0.0;
	pnode->coor.n = 0.0;
	pnode->next = NULL;
	pnode->left = NULL;
	pnode->right = NULL;
	return;
}


//
// peak_node_pool_head -> chunk0
//                            |
//                           node_array[initial_tetra_chunk_size]
//
int allocate_peak_nodes(void)
{
	struct peak_node_chunk * new_chunk;
	struct peak_node * new_node_array;
	int n_ct;
	
	if (get_param_safe_limit(&my_param) < get_param_init_size(&my_param) + 
	    peak_node_pool_head.chunk_ct * get_param_incre_size(&my_param)) {
		fprintf(STDERR, "allocate_peak_nodes() : ERROR  num of peak nodes is over the safe limit (=%d)\n",
			get_param_safe_limit(&my_param));
		fprintf(STDERR, "allocate_peak_nodes() : ERROR      init_size = %d\n", get_param_init_size(&my_param));
		fprintf(STDERR, "allocate_peak_nodes() : ERROR      chunk_ct = %d\n", peak_node_pool_head.chunk_ct);
		fprintf(STDERR, "allocate_peak_nodes() : ERROR      incre_size = %d\n", get_param_incre_size(&my_param));
		return FAIL;
	}
	new_chunk = (struct peak_node_chunk *) malloc(sizeof(struct peak_node_chunk));
	if (new_chunk == NULL) {
		fprintf(STDERR, 
			"allocate_peak_nodes() : ERROR allocation failed (new_chunk)\n");
		return FAIL;
	}
	initialize_pnode_chunk(new_chunk);
	if (peak_node_pool_head.next == NULL) {
		new_node_array = (struct peak_node *) malloc(get_param_init_size(&my_param) * sizeof(struct peak_node));
		if (new_node_array == NULL) {
			fprintf(STDERR, 
				"allocate_peak_node() : ERROR initial allocation failed (new_node_array[%d])\n",
				get_param_init_size(&my_param));
			return FAIL;
		}
		new_chunk->array_size = get_param_init_size(&my_param);
	} else {
		new_node_array = (struct peak_node *) malloc(get_param_incre_size(&my_param) * sizeof(struct peak_node));
		if (new_node_array == NULL) {
			fprintf(STDERR, 
				"allocate_search_node() : ERROR incremental allocation failed (new_node_array[%d])\n",
				get_param_incre_size(&my_param));
			return FAIL;
		}
		new_chunk->array_size = get_param_incre_size(&my_param);
	}
	for (n_ct = 0; n_ct < get_param_init_size(&my_param); n_ct++)
		initialize_peak_node(new_node_array + n_ct);
	new_chunk->node_array = new_node_array;
	put_pnode_chunk(&peak_node_pool_head, new_chunk);
	return SUCCESS;
}


//
// list_head -> NULL
//
void set_up_empty_pnode_list(struct peak_node * list_head)
{
	initialize_peak_node(list_head);
	return;
}


//
// list_head -> ... -> (top_chunk->p_node1) -> (top_chunk->p_node0) -> old_pnodes
//
void fill_peak_node_list(struct peak_node * list_head)
{
	struct peak_node_chunk * top_chunk;
	int i;
	
	top_chunk = peak_node_pool_head.next;
	for (i=0; i < top_chunk->array_size ; i++)
		put_peak_node(list_head, top_chunk->node_array + i);
	return;
}


struct peak_node * assign_peak_node(void)
{
	struct peak_node * pnode;
	
	pnode = take_peak_node(&unused_peak_node_head);
	if (pnode == NULL) {
		if (DEBUG) fprintf(STDERR, "assign_peak_node() : DEBUG allocate peak_nodes (current chunk_ct=%d)\n", peak_node_pool_head.chunk_ct);
		if (allocate_peak_nodes() == FAIL) {
			fprintf(STDERR, "assign_peak_node() : ERROR allocation failed\n");
			exit(1);
		}
		fill_peak_node_list(&unused_peak_node_head);
		pnode = take_peak_node(&unused_peak_node_head);
	}
	return pnode;
}


void recall_peak_node(struct peak_node * pnode)
{
	initialize_peak_node(pnode);
	put_peak_node(&unused_peak_node_head, pnode);
	return;
}


//
//  {a0, b0, c0} > {a1, b1, c1}
//  <=> a0 > a1,
//      or (a0 == a1) && (b0 > b1)
//      or (a0 == a1) && (b0 == b1) && (c0 > c1)
//
HNV_COMPARE_RESULT which_peak_is_greater(struct peak_node * current_node,
				      struct peak_node * new_node)
{
	if (current_node->coor.k > new_node->coor.k) {
		return HNV_LEFT;
	} else if (current_node->coor.k == new_node->coor.k) {
		if (current_node->coor.l > new_node->coor.l) {
			return HNV_LEFT;
		} else if (current_node->coor.l == new_node->coor.l) {
			if (current_node->coor.m > new_node->coor.m) {
				return HNV_LEFT;
			} else if (current_node->coor.m == new_node->coor.m) {
				if (current_node->coor.n > new_node->coor.n) {
					return HNV_LEFT;
				} else if (current_node->coor.n == new_node->coor.n) {
					return HNV_EQUAL;
				} else {
					return HNV_RIGHT;
				}
			} else {
				return HNV_RIGHT;
			}
		} else {
			return HNV_RIGHT;
		}
	} else {
		return HNV_RIGHT;
	}
}


void print_peak_node_chunk(struct peak_node_chunk * p_chunk)
{
	fprintf(STDOUT, "print_peak_node_chunk() : chunk = %x, array_size = %d, chunk_ct = %d, next chunk = %x\n",
		(unsigned int) p_chunk, p_chunk->array_size, p_chunk->chunk_ct, (unsigned int) p_chunk->next);
	return;
}


void print_pnode_chunk_list(struct peak_node_chunk * chunk)
{
	struct peak_node_chunk * next_chunk;
	int c_ct;
	
	next_chunk = chunk;
	c_ct = 0;
	while (next_chunk != NULL) {
		fprintf(STDOUT, "print_pnode_chunk_list() : start of %d-th peak node chunk\n", c_ct);
		print_peak_node_chunk(next_chunk);
		next_chunk = next_chunk->next;
		c_ct++;
	}
	return;
}


void print_peak_node(struct peak_node * pnode)
{
	fprintf(STDOUT, "print_peak_node() : locator = { %8.3f, %8.3f, %8.3f, %8.3f }, &pnode = %x\n",
		pnode->coor.k, pnode->coor.l, pnode->coor.m, pnode->coor.n, (unsigned int) &pnode);
	fprintf(STDOUT, "print_peak_node() : pnode->left = %x, pnode->right = %x, pnode->next = %x\n",
		(unsigned int) pnode->left, (unsigned int) pnode->right, (unsigned int) pnode->next);
	return;
}


void walk_peak_node_tree(struct peak_node * pnode)
{
	if (pnode != NULL) {
		walk_peak_node_tree(pnode->left);
		print_peak_node(pnode);
		walk_peak_node_tree(pnode->right);
	}
	return;
}


void print_peak_node_tree(void)
{
	fprintf(STDOUT, "print_peak_node_tree() : START\n");
	walk_peak_node_tree(peak_tree_root);
	fprintf(STDOUT, "print_peak_node_tree() : END\n");
	fprintf(STDOUT, "\n");
	return;
}


void print_out_new_peaks(FILE * stream, struct peak_node * peak)
{
	struct peak_node * next_peak;
	int p_ct;
	
	next_peak = peak->next;
	p_ct = 0;
	while (next_peak != NULL) {
		next_peak = next_peak->next;
		p_ct++;
	}
	fprintf(stream, "\n");
	fprintf(stream, "<ADDED PEAKS    num=\"%d\">\n", p_ct);
	next_peak = peak->next;
	p_ct = 0;
	while (next_peak != NULL) {
		fprintf(stream, "<peak> [%4d] coor = { %8.3f, %8.3f, %8.3f, %8.3f }\n",
			p_ct, next_peak->coor.k, next_peak->coor.l, 
			next_peak->coor.m, next_peak->coor.n);
		next_peak = next_peak->next;
		p_ct++;
	}
	fprintf(stream, "</ADDED PEAKS>\n");
	fprintf(stream, "\n");

	return;
}


void print_pnode_list(struct peak_node * peak)
{
	struct peak_node * next_peak;
	int p_ct;
	
	next_peak = peak->next;
	p_ct = 0;
	while (next_peak != NULL) {
		fprintf(STDOUT, "print_pnode_list() : start of %d-th peak node\n", p_ct);
		print_peak_node(next_peak);
		next_peak = next_peak->next;
		p_ct++;
	}
	fprintf(STDOUT, "\n");
	return;
}


void initialize_peak_node_tree(void)
{
	initialize_pnode_chunk(&peak_node_pool_head);
	if (allocate_peak_nodes() == FAIL) {
		fprintf(STDERR, "initialize_peak_node_pool() : ERROR allocation failed\n");
		exit(1);
	}
	set_up_empty_pnode_list(&unused_peak_node_head);
	fill_peak_node_list(&unused_peak_node_head);
	peak_tree_root = NULL;
	return;
}


void finalize_peak_node_tree(void)
{
	struct peak_node_chunk * next_chunk;
	struct peak_node_chunk * save_chunk;
	
	next_chunk = peak_node_pool_head.next;
	while (next_chunk != NULL) {
		save_chunk = next_chunk;
		next_chunk = next_chunk->next;
		free(save_chunk);
	}
	peak_tree_root = NULL;
	return;
}


//
//        peak_tree_root =
//                ------node0------
//                |               |
//            --node1--        --node2--
//            |       |        |       |
//         node11   node12    node21  node22
//
void register_in_peak_tree(struct peak_node * new_pnode)
{
	struct peak_node * current_pnode;
	struct peak_node * old_pnode;
	HNV_COMPARE_RESULT comp_flag;
	
	if (peak_tree_root != NULL) {
		current_pnode = peak_tree_root;
		while (current_pnode != NULL) {
			old_pnode = current_pnode;
			comp_flag = which_peak_is_greater(current_pnode, new_pnode);
			if (comp_flag == HNV_LEFT) {	// current_pnode > new_pnode
				current_pnode = current_pnode->left;
			} else if (comp_flag == HNV_RIGHT) {
				current_pnode = current_pnode->right;
			} else {
				fprintf(STDERR,
					"register_in_peak_tree() : ERROR peak {%8.3f, %8.3f, %8.3f, %8.3f} is already registered (new_pnode=%x)\n",
					new_pnode->coor.k, new_pnode->coor.l, new_pnode->coor.m, new_pnode->coor.n, (unsigned int) new_pnode);
				fprintf(STDERR,
					"register_in_peak_tree() : ERROR the peak of cur_pnode %x is {%8.3f, %8.3f, %8.3f, %8.3f}\n",
					(unsigned int) current_pnode, current_pnode->coor.k, current_pnode->coor.l, current_pnode->coor.m, current_pnode->coor.n);
				exit(1);
			}
		}
		if (comp_flag == HNV_LEFT) {	// old_snode > new_snode
			old_pnode->left = new_pnode;
		} else {
			old_pnode->right = new_pnode;
		}
	} else {
		peak_tree_root = new_pnode;
	}
	return;
}


struct peak_node * check_if_peak_is_processed(struct peak_node * new_pnode)
{
	struct peak_node * current_pnode;
	HNV_COMPARE_RESULT comp_flag;
	
	current_pnode = peak_tree_root;
	while (current_pnode != NULL) {
		comp_flag = which_peak_is_greater(current_pnode, new_pnode);
		if (comp_flag == HNV_LEFT) {	// current_snode > new_snode
			current_pnode = current_pnode->left;
		} else if (comp_flag == HNV_RIGHT) {
			current_pnode = current_pnode->right;
		} else {
			return current_pnode;	// processed
		}
	}
	return (struct peak_node *) NULL;	// not processed
}


// "bottom vertex" - { 1.0, 1.0, 1.0, 1.0 }
// = {(a+b+c)/4, (-a+b-c)/4, (-a-b+c)/4, (a-b-c)/4} + {h, h, h, h} -  { 1.0, 1.0, 1.0, 1.0 }
void set_new_peak(struct peak_node * new_peak, struct hnv_coor_3d * pt_bottom, float height)
{
	new_peak->coor.k = (  pt_bottom->a + pt_bottom->b + pt_bottom->c) / 4.0 + height - 1.0;
	new_peak->coor.l = (- pt_bottom->a + pt_bottom->b - pt_bottom->c) / 4.0 + height - 1.0;
	new_peak->coor.m = (- pt_bottom->a - pt_bottom->b + pt_bottom->c) / 4.0 + height - 1.0;
	new_peak->coor.n = (  pt_bottom->a - pt_bottom->b - pt_bottom->c) / 4.0 + height - 1.0;
	return;
}


// projection: {k, l, m, n} -->==> {k-l-m+n, k+l-m-n, k-l+m-n}
void compute_top_vertex(struct hnv_coor_3d * vertex, struct peak_node * peak, int top_ind)
{	
	vertex->a = peak->coor.k - peak->coor.l - peak->coor.m + peak->coor.n
		    + neighbours[top_ind][0];
	vertex->b = peak->coor.k + peak->coor.l - peak->coor.m - peak->coor.n
		    + neighbours[top_ind][1];
	vertex->c = peak->coor.k - peak->coor.l + peak->coor.m - peak->coor.n
		    + neighbours[top_ind][2];
	return;
}


void compute_bottom_vertex(struct hnv_coor_3d * bottom, struct hnv_coor_3d * top,
		      int bottom_ind)
{
	bottom->a = top->a + bottoms[bottom_ind][0];
	bottom->b = top->b + bottoms[bottom_ind][1];
	bottom->c = top->c + bottoms[bottom_ind][2];
	return;
}


#define IS_NOT_BOUNDARY 0
#define IS_BOUNDARY 1
// BD-check:  if "top is out" && "bottom is in"
int check_if_on_boundary(struct hnv_coor_3d * top, struct hnv_coor_3d * bottom, 
			 struct hnv_cone * std_cone, struct hnv_cone * conj_cone)
{
	float height_std_top;
	float height_std_bottom;
	float height_conj_top;
	float height_conj_bottom;
	
	height_std_top = height_std_cone(std_cone, top);
	height_conj_top = height_conj_cone(conj_cone, top);
	height_std_bottom = height_std_cone(std_cone, bottom);
	height_conj_bottom = height_conj_cone(conj_cone, bottom);
	
	if ( (height_std_top < height_conj_top)
	     && (height_std_bottom > height_conj_bottom)) {
		return IS_BOUNDARY;
	} else {
		return IS_NOT_BOUNDARY;
	}
}


int compute_bd_peak_list(struct hnv_cone * std_cone, struct hnv_cone * conj_cone)
{
	struct hnv_coor_3d cur_top;
	struct hnv_coor_3d cur_bottom;
	struct peak_node * new_peak;
	struct peak_node * last_peak;
	struct peak_node * next_peak;
	float height_std_bottom;
	int top_ind, bottom_ind;
	int bd_peak_list_length;
	int check_flag;
	
	bd_peak_list_length = 0;
	// Get top vertices of the current BD tetras from "bd_tetra_list_head"
	{
		struct tetrahedron * next_tetra;
		int tetra_ct;
		
		tetra_ct = get_tetra_tetra_ct(&bd_tetra_list_head);
		next_tetra = get_tetra_next_tetra(&bd_tetra_list_head);
		last_peak = &bd_peak_list_head;
		while (next_tetra != NULL) {
			if (bd_peak_list_length >= tetra_ct) {
				fprintf(STDERR, "compute_bd_peak_list() : ERROR overflow  next_tetra = %x, bd_peak_list_length (=%d) >= tetra_ct (=%d)\n",
					(unsigned int) next_tetra, bd_peak_list_length, tetra_ct);
				exit(1);
			}
			get_tetra_bottom_vertex(&cur_bottom, next_tetra);
			height_std_bottom = height_std_cone(std_cone, &cur_bottom);
			new_peak = assign_peak_node();
			set_new_peak(new_peak, &cur_bottom, height_std_bottom);
			if (check_if_peak_is_processed(new_peak) != NULL) {
				recall_peak_node(new_peak);
			} else {
				if (0) {
					fprintf(STDOUT, "compute_bd_peak_list() : DEBUG cur_bottom = { %8.3f, %8.3f, %8.3f }, height = %8.3f\n",
						cur_bottom.a, cur_bottom.b, cur_bottom.c, height_std_bottom);
					fprintf(STDOUT, "compute_bd_peak_list() : DEBUG new_peak = %x\n", (unsigned int) new_peak);
					print_peak_node(new_peak);
				}
				register_in_peak_tree(new_peak);
				put_peak_node(last_peak, new_peak);
				bd_peak_list_length++;
				last_peak = new_peak;
			}
			next_tetra = get_tetra_next_tetra(next_tetra);
		}
	}
	
	next_peak = &bd_peak_list_head;
	while (next_peak != NULL) {
		for (top_ind = 0; top_ind < 12; top_ind++) {
			compute_top_vertex(&cur_top, next_peak, top_ind);
			for (bottom_ind = 0; bottom_ind < 4; bottom_ind++) {
				compute_bottom_vertex(&cur_bottom, &cur_top, bottom_ind);
				check_flag = check_if_on_boundary(&cur_top,
						&cur_bottom, std_cone,  conj_cone);
				if ( check_flag == IS_BOUNDARY) {
					height_std_bottom = height_std_cone(std_cone, &cur_bottom);
					new_peak = assign_peak_node();
					set_new_peak(new_peak, &cur_bottom, height_std_bottom);
					if (check_if_peak_is_processed(new_peak) != NULL) {
						recall_peak_node(new_peak);
					} else {
						if (0) {
							fprintf(STDOUT, "compute_bd_peak_list() : DEBUG cur_bottom = { %8.3f, %8.3f, %8.3f }, height = %8.3f\n",
								cur_bottom.a, cur_bottom.b, cur_bottom.c, height_std_bottom);
							fprintf(STDOUT, "compute_bd_peak_list() : DEBUG new_peak = %x\n", (unsigned int) new_peak);
							print_peak_node(new_peak);
						}
						register_in_peak_tree(new_peak);
						put_peak_node(last_peak, new_peak);
						bd_peak_list_length++;
						last_peak = new_peak;
					}
				}
			}
		}
		next_peak = next_peak->next;
	}
	return bd_peak_list_length;
}


struct hnv_cone * closure_of_std_cone(struct hnv_cone * std_cone, struct hnv_cone * conj_cone)
{
	struct hnv_cone * new_std_cone;
	int bd_peak_ct;
	int cone_size;
	
	bd_peak_ct = compute_bd_peak_list(std_cone, conj_cone);
	if (0) {
		fprintf(STDERR, "closure_of_std_cone() : Add %d peaks to the standard cone (closure)\n", bd_peak_ct);
	}
	if (1) {
		print_out_new_peaks(STDOUT, &bd_peak_list_head);
	}
	
	cone_size = get_cone_peak_size(std_cone) + bd_peak_ct;
	new_std_cone = allocate_hnv_cone(get_cone_dim(std_cone), cone_size);
	if (new_std_cone == NULL) {
		fprintf(STDERR, "closure_of_std_cone() : ERROR allocation failed (new_std_cone[%d])\n",
			 cone_size);
		exit(1);
	}	
	
	{
		struct peak_node * next_peak;
		int p_ct;
		int sum_mod;
		
		// Put original peaks
		copy_hnv_cone(new_std_cone, std_cone);
		for (p_ct = 0; p_ct < get_cone_peak_size(std_cone); p_ct++)
			set_peak_redundance(new_std_cone, p_ct, HNV_UNKNOWN_P);
	
		// Add new peaks
		next_peak = bd_peak_list_head.next;
		while (next_peak != NULL) {
			set_peak_coor(new_std_cone, p_ct, &(next_peak->coor));
			set_peak_redundance(new_std_cone, p_ct, HNV_UNKNOWN_P);
			sum_mod = next_peak->coor.k + next_peak->coor.l 
				+ next_peak->coor.m + next_peak->coor.n;
			sum_mod = sum_mod % (get_cone_dim(std_cone) - 1);
			set_peak_sum_mod3(new_std_cone, p_ct, sum_mod);
			next_peak = next_peak->next;
			p_ct++;
		}
		set_cone_peak_ct(new_std_cone, p_ct);
	}
	return new_std_cone;
}


//===============================================
// Public methods
//===============================================

void compute_closure(void)
{
	struct hnv_cone * new_std_cone;
	
	if(1) {
		fprintf(STDOUT, "HNV> Take closure of the current flow\n");
	}
	initialize_peak_node_tree();
	set_up_empty_pnode_list(&bd_peak_list_head);
	new_std_cone = closure_of_std_cone(my_standard_cone, my_conjugate_roof);
	finalize_peak_node_tree();
	free_cones_and_tetras();
	
	my_standard_cone = new_std_cone;
	if (get_remove_redundant_peaks() == HNV_TRUE)
		remove_redundant_peaks(my_standard_cone);
	if (0) {
		fprintf(STDOUT, "compute_closure() : DEBUG  START of my_standard_cone\n");
		print_hnv_cone_short(my_standard_cone);
		fprintf(STDOUT, "compute_closure() : DEBUG  END of my_standard_cone\n");
		fprintf(STDOUT, "\n");
	}
	my_standard_roof = compute_roof(my_standard_cone);
	if (0) {
		fprintf(STDOUT, "compute_closure() : DEBUG  START of my_standard_roof\n");
		print_hnv_cone_short(my_standard_roof);
		fprintf(STDOUT, "compute_closure() : DEBUG  END of my_standard_roof\n");
		fprintf(STDOUT, "\n");
	}
	my_conjugate_roof = compute_3d_case(my_standard_cone);
	if (0) {
		fprintf(STDOUT, "compute_closure() : DEBUG  START of my_conjugate_roof\n");
		print_hnv_cone_short(my_conjugate_roof);
		fprintf(STDOUT, "compute_closure() : DEBUG  END of my_conjugate_roof\n");
		fprintf(STDOUT, "\n");
	}
	remove_redundant_peaks(my_standard_cone);
	if (0) {
		if (get_remove_redundant_peaks() != HNV_TRUE) {
			fprintf(STDOUT, "compute_closure() : DEBUG  START of my_standard_cone (reduced)\n");
			print_hnv_cone_short(my_standard_cone);
			fprintf(STDOUT, "compute_closure() : DEBUG  END of my_standard_cone (reduced)\n");
			fprintf(STDOUT, "\n");
		}
	}
	if (0) {
		fprintf(STDOUT, "compute_closure() : DEBUG START of closed_traj_list_head\n");
		print_traj_list(&closed_traj_list_head);
		fprintf(STDOUT, "compute_closure() : DEBUG END of closed_traj_list_head\n");
		fprintf(STDOUT, "compute_closure() :\n");
		fprintf(STDOUT, "compute_closure() : DEBUG START of open_traj_list_head\n");
		print_traj_list(&open_traj_list_head);
		fprintf(STDOUT, "compute_closure() : DEBUG END of open_traj_list_head\n");
		fprintf(STDOUT, "compute_closure() :\n");
		fprintf(STDOUT, "compute_closure() : DEBUG START of bd_tetra_list_head\n");
		print_tetra_list(&bd_tetra_list_head);
		fprintf(STDOUT, "compute_closurew() : DEBUG END of bd_tetra_list_head\n");
		fprintf(STDOUT, "\n");
	}
	
	if (check_tflow_seq_array(&my_closure_tetrahedron_flow) == SUCCESS) {
		free_tetra_flow(&my_closure_tetrahedron_flow);
	}
	set_up_tetra_flow(&my_closure_tetrahedron_flow);
	return;
}

